/////////////////////////////////////////////////////////////////////////////
//
// wxWebServer - a simple web server using wxWidgets framework
//
// File:        requesthandler.cpp
// 
// Author:      Fedja Stevanovic
// 
// Copyright:   Copyright (C) 2007 Fedja Stevanovic (tutubibi.net)
//
// Licence:
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#if defined(__GNUG__) && !defined(NO_GCC_PRAGMA)
#pragma implementation "requesthandler.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <wx/string.h>
#include <wx/mimetype.h>
#include <wx/filename.h>
#include <wx/file.h>
#include <wx/ffile.h>
#include <wx/tokenzr.h>
#include <wx/arrstr.h>
#include <wx/utils.h>
#include <wx/process.h>
#include <wx/socket.h>
#include "requesthandler.h"
#include "contenttypes.h"

// 5000 should be enough for all the headers!
#define BUFSIZE 5000

#define METHOD_GET  1
#define METHOD_POST 2
#define METHOD_HEAD 3

#define SERVER_NAME	"wxWebServer-1.0.0"


/*!
* RequestHandler constructors
 */

CRequestHandler::CRequestHandler(wxSocketBase* Socket, const wxString& wwwRoot, const wxString& defaultFile,
								 const wxString& cgiExt, const wxString& cgiCmd, const wxString& assumedContentType,
									wxString* reqLog, wxString* resLog )
{
    m_Socket = Socket;
	m_wwwRoot = wwwRoot;
	m_defaultFile = defaultFile;
	m_reqBody = NULL;
	m_cgiExt = cgiExt;
	m_cgiHandler = cgiCmd;
	m_assumedContentType = assumedContentType;
	m_reqLog = reqLog;
	m_resLog = resLog;
	
	wxSockAddress* localInfo = new wxIPV4address();
	wxSockAddress* peerInfo = new wxIPV4address();
	//wxIPV4address localInfo, peerInfo;
	if( m_Socket->GetLocal(*localInfo) )
	{
		m_localPort.Printf(_T("%d"), ((wxIPV4address*)localInfo)->Service() );
		m_localHost = ((wxIPV4address*)localInfo)->Hostname();
	} else {
		m_localPort = wxEmptyString;
		m_localHost = wxEmptyString;
	}
	
	if( m_Socket->GetPeer(*peerInfo) )
	{
		m_peerPort.Printf(_T("%d"), ((wxIPV4address*)peerInfo)->Service() );
		m_peerHost = ((wxIPV4address*)peerInfo)->Hostname();
	} else {
		m_peerPort = wxEmptyString;
		m_peerHost = wxEmptyString;
	}
	
	// allocated receive buffer
	buf = new char[BUFSIZE];
	
	Create();
    Run();
}


/*!
* CRequestHandler destructor
 */

CRequestHandler::~CRequestHandler()
{
	if( m_reqBody != NULL )
		delete [] m_reqBody;
	
	delete [] buf;
}

/*!
* ReturnError 
 */

void* CRequestHandler::ReturnError(int code, char* description)
{
	char response[500];
	sprintf( response, "HTTP/1.1 %ld %s\r\nserver: wxWebServer-1.0.0\r\ncontent-type: text/plain\r\ncontent-length: %ld\r\n\r\n%s", code, description, strlen(description), description );
	m_Socket->Write(response, strlen(response));
	m_Socket->Destroy();
	return NULL;
}

/*!
* Parse request headers
 */
bool CRequestHandler::ParseHeaders()
{
	wxStringTokenizer tkz( m_reqFullHeaders, wxT("\r\n"));
	while ( tkz.HasMoreTokens() )
	{
		wxString token = tkz.GetNextToken();
		// token is a header in the format "name: value"
		int i = token.Find(_T(":"));
		if( i == wxNOT_FOUND )
			return false;
		wxString name = token.Mid(0, i);
		wxString value = token.Mid(i+1);
		name.Trim();
		name.MakeLower();
		value.Trim();
		
		m_reqHeaders[name] = value;
	}
	
	return true;
	
}


/*!
* Entry will process the request
 */

void* CRequestHandler::Entry()
{
	// read request
	char* headers = NULL;
	char* method = NULL;
	char* version = NULL;
	char* url = NULL;
	char* eoh = NULL;
	char* pos;
	wxUint32 count; 
	
	try {
		
		// Read the data
		m_Socket->Read(buf, BUFSIZE);
		count = m_Socket->LastCount();
		eoh = strstr(buf, "\r\n\r\n");
		if( eoh == NULL ) {
			return ReturnError(400, "Bad Request");
		}
		
		// find method
		method = buf;
		pos = strstr(method," ");
		if ( pos == NULL ) {
			return ReturnError(400, "Bad Request");
		}
		url = pos+1;
		if( url[0] != '/' )
			return ReturnError(400, "Bad Request");
		*pos = 0x00;
		
		// find uri
		pos = strstr(url," ");
		if ( pos == NULL )	{
			return ReturnError(400, "Bad Request");
		}
		version = pos+1;
		*pos = 0x00;
		
		// find version
		pos = strstr(version,"\r");
		if ( pos == NULL )	{
			return ReturnError(400, "Bad Request");
		}
		headers = pos+2;
		*pos = 0x00;
		
		// mark the end of headers, so each is separated by a single CRLF
		eoh += 2;
		*eoh = 0x00;
		eoh += 2;
		
		m_reqMethod = wxString( method, wxConvLibc );
		m_reqUrl = wxString( url, wxConvLibc );
		m_reqVersion = wxString( version, wxConvLibc );
		m_reqFullHeaders = wxString( headers, wxConvLibc );
		m_reqFullUrl = m_reqUrl;
		
		int qspos = m_reqUrl.Find(_T("?")); // has query string?
		if( qspos != wxNOT_FOUND )  
		{
			m_queryString = m_reqUrl.Mid( qspos+1 );
			m_reqUrl.Remove(qspos);
		} else
			m_queryString = wxEmptyString;
		
		
		if( !ParseHeaders() )
			return ReturnError(400, "Bad Request");
		
		if( m_reqUrl.Last() == '/' ) // default file
			m_reqUrl.append( m_defaultFile );
		
		if( !strcmp( method, "GET" ) )
			m_methodCode = METHOD_GET;
		else if ( !strcmp( method, "HEAD" ) )
			m_methodCode = METHOD_GET;
		else if( !strcmp( method, "POST" ) )
		{
			m_methodCode = METHOD_POST;
			// POST, has a body ... receive more
			unsigned long hl = eoh - method; // header length
			wxString strCL = m_reqHeaders[_T("content-length")];
			if( !strCL.ToULong(&m_reqContentLength) ) {
				return ReturnError(400, "Bad Request");
			} else {
				m_reqBody = new char[m_reqContentLength];
				if( count < BUFSIZE ) { 	// all received
					memcpy(m_reqBody, eoh, m_reqContentLength);
				}
				else { // need more to download
					memcpy(m_reqBody, eoh, count-hl);
					int i = count-hl;
					void* next = m_reqBody + i;
					// read the rest of the data
					m_Socket->Read(next, m_reqContentLength);
				}
			}
		} 
		else
			return ReturnError(405, "Method Not Allowed"); // unsopported method
		
		m_fullPath = m_wwwRoot;
		// based on the url lets figure out the filename + path
		m_fullPath.append( m_reqUrl );
		
		// is it PHP or static resouce
		if( m_reqUrl.EndsWith( m_cgiExt) )
		{
			DoCGI();
		}
		
		// it's not CGI, so it should be a static file from the file system
		
		wxFileName fname;
		fname.Assign( m_fullPath );

		if( !fname.FileExists() )
			return ReturnError(404, "Not Found");

		wxString contentType = ContentTypes::getInstance()->getContentType(fname.GetExt());
		
		if( contentType.IsEmpty() )
			contentType = ContentTypes::getInstance()->getFallbackContentType();
		
		wxFile responseFile;
		if( responseFile.Open( m_fullPath, wxFile::read ) )
		{
			long fileSize = responseFile.Length();
			char* responseBuffer = new char[fileSize];
			long contentLength = responseFile.Read( responseBuffer, fileSize );
			if( contentLength > 0 ) {
				// let's reuse buf, no need for request data at this point
				sprintf(buf, "HTTP/1.1 200 OK\r\nserver: wxWebServer-1.0.0\r\ncontent-type: %s\r\ncontent-length: %ld\r\n\r\n", (char*)contentType.char_str(), contentLength );
				long len = strlen(buf);
				m_Socket->Write(buf, len );
				// send body only if not HEAD request
				if( m_methodCode != METHOD_HEAD )
					m_Socket->Write(responseBuffer, contentLength );
			}
			else
				return ReturnError(500, "Server Error");
		}
		else 
			return ReturnError(404, "Not Found");
		
		// We are done with the socket, destroy it
		m_Socket->Destroy();
		
	} catch (int e)  {
		return ReturnError(500, "Server Error");
	}
	
	return NULL;
}

/*!
* DoCGI
 */

void* CRequestHandler::DoCGI()
{
	try {
		
		// disable all logging to eliminate error popups that occur sometimes on wxExecute
		wxLogNull logNo;
		
		wxSetEnv( _T("DOCUMENT_ROOT"), m_wwwRoot );
		wxSetEnv( _T("GATEWAY_INTERFACE"), _T("CGI/1.1") );
		
		wxSetEnv( _T("REQUEST_METHOD"), m_reqMethod );
		wxSetEnv( _T("REQUEST_URI"), m_reqFullUrl );
		wxSetEnv( _T("SCRIPT_FILENAME"), m_fullPath );
		wxSetEnv( _T("SCRIPT_NAME"), m_reqUrl );
		wxSetEnv( _T("QUERY_STRING"), m_queryString );

		wxSetEnv( _T("PATH_INFO"), m_fullPath );

		wxSetEnv( _T("SERVER_NAME"), m_localHost );
		wxSetEnv( _T("SERVER_PORT"), m_localPort );
		wxSetEnv( _T("SERVER_PROTOCOL"), _T("HTTP/1.1") );
		wxSetEnv( _T("SERVER_SOFTWARE"), _T(SERVER_NAME) );
		
		wxSetEnv( _T("REMOTE_ADDR"), m_peerHost );
		wxSetEnv( _T("REMOTE_PORT"), m_peerPort );
		
		wxSetEnv( _T("HTTP_CONNECTION"), m_reqHeaders[_T("connection")] );
		wxSetEnv( _T("HTTP_COOKIE"), m_reqHeaders[_T("cookie")] );
		wxSetEnv( _T("HTTP_HOST"), m_reqHeaders[_T("host")] );
		wxSetEnv( _T("HTTP_USER_AGENT"), m_reqHeaders[_T("user-agent")] );
		
		wxString cmd = m_cgiHandler + _T(" ") + m_fullPath;
		
		if( m_methodCode != METHOD_POST )
		{

			wxUnsetEnv( wxString(_T("CONTENT_LENGTH")) );
			wxUnsetEnv( wxString(_T("CONTENT_TYPE")) );
		} 
		else 
		{ 
			wxSetEnv( wxString(_T("CONTENT_LENGTH")), m_reqHeaders[_T("content-length")] );
			wxSetEnv( wxString(_T("CONTENT_TYPE")), m_reqHeaders[_T("content-type")] );
		}
		
		wxFile* postInputFile = NULL;
		wxString postInputFName;
		wxFile* postOutputFile = NULL;
		wxString postOutputFName;
		char* fbuf;

		try {

			wxFileName ofname, ifname;

			if( m_methodCode == METHOD_POST ) {
				postInputFile = new wxFile();
				postInputFName = ifname.CreateTempFileName(_("wxws_"), postInputFile );
				if( postInputFName.IsEmpty() )
					return ReturnError( 510, "Server Error on CGI Preparation");
				postInputFile->Write( m_reqBody, m_reqContentLength );
				postInputFile->Close();

				cmd.Append(_(" <"));
				cmd.Append(postInputFName);
			}

			postOutputFile = new wxFile();
			postOutputFName = ifname.CreateTempFileName(_("wxws_"), postOutputFile );
			postOutputFile->Close();

			cmd.Append(_(" >"));
			cmd.Append(postOutputFName);

			// Using wxShell proved to much more reliable than wxExecute or wxProcess::Open and redirect 
			// Of course, using native CreateProcess or exec functions  would be much better 
			// but this is wxWidgets demo as well so we stick with wxShell
			if( wxShell(cmd) )
			{
				wxFile responseFile;

				responseFile.Open( postOutputFName, wxFile::read );
				size_t flen = responseFile.Length();
				fbuf = new char[flen];
				responseFile.Read( fbuf, flen );
				responseFile.Close();

				if( !m_assumedContentType.IsEmpty() ) {
					// assume content type
					wxString headers;
					headers.Append(_T("HTTP/1.1 200 OK\r\nserver: wxWebServer-1.0.0\r\ncontent-type: %s\r\ncontent-length: %ld\r\n\r\n"));
					// let's reuse buf, no need for request data at this point
					sprintf(buf, (char*)headers.char_str(), (char*) m_assumedContentType.char_str(), flen );
					long len = strlen(buf);
					m_Socket->Write(buf, len );
					m_Socket->Write(fbuf, flen );								
				} else {

					int offset = 4;
					char* pos = strstr(fbuf, "\r\n\r\n"); // WINDOWS
					if( pos == NULL ) {
						offset = 2;
						pos = strstr(fbuf, "\n\n");  // UNIX
						if( pos == NULL ) {
							pos = strstr(fbuf, "\r\r");  // OLD MAC
							if( pos == NULL )
								throw 500;
						}
					}

					// everything OK with headers
					*pos = 0x00;
					pos += offset;

					size_t conlen = flen - (pos - fbuf);

					wxString headers = wxString(fbuf, wxConvLibc);

					size_t start = 0;
					if( !headers.StartsWith(_T("HTTP/")) )
					{
						headers.Prepend(_T("HTTP/1.1 200 OK\r\n"));
					} 

					//headers.Append(_T("Set-Cookie: wxWebServerCookie=SetForTesting\r\n"));
					headers.Append(_T("\r\nserver: wxWebServer-1.0.0\r\ncontent-length: %ld\r\n\r\n"));

					// let's reuse buf, no need for request data at this point
					sprintf(buf, (char*)headers.char_str(), conlen );
					long len = strlen(buf);
					m_Socket->Write(buf, len );
					m_Socket->Write(pos, conlen );
				}

				m_Socket->Destroy();
				delete [] fbuf;

			} 

			wxRemoveFile( postOutputFName );
			if( m_methodCode == METHOD_POST ) 
				wxRemoveFile( postInputFName );	

		} catch (int e) {
			if( m_methodCode == METHOD_POST ) {
				if( postInputFile != NULL )
					delete postInputFile;
				wxRemoveFile( postInputFName );	
			}
			if( postOutputFile != NULL )
				delete postOutputFile;
			wxRemoveFile( postOutputFName );
			if( fbuf )
				delete [] fbuf;	
			if( e == 500 )
				return ReturnError(500, "Server Error - No headers returned");
			return ReturnError(500, "Server Error on CGI Processing");
		}
		
	} catch (int e)  {
		return ReturnError(500, "Server Error");
	}
	
	return ReturnError( 511, "Something bad happened");
}
